/*
 * led_prog.c
 *
 *  Created on: Aug 3, 2023
 *      Author: true
 *
 */


#include <stdint.h>
#include "hsv2rgb.h"
#include "led_sk6x_spi.h"
#include "prng.h"



void ledprog0_twinkle_white();
void ledprog1_twinkle_rgb_rand();
void ledprog2_twinkle_rgb_set();
void ledprog3_rainbow();
void ledprog4_flicker_same();
void ledprog5_flicker_all();
void ledprog6_ramper();


static uint8_t settings[16][4];

const void (*led_prog_list[])() = {
	ledprog0_twinkle_white,
	ledprog1_twinkle_rgb_rand,
	ledprog2_twinkle_rgb_set,
	ledprog3_rainbow,
	ledprog4_flicker_same,
	ledprog5_flicker_all,
	ledprog6_ramper
};

void (*ledprog)();
static uint32_t work[4];

static uint8_t r, g, b;



void ledprog_default()
{
	settings[3][2] = 2;
	settings[3][3] = 30;

	settings[4][1] = 248;
	settings[4][2] = 170;
	settings[4][3] = 16;

	settings[6][1] = 192;
	settings[6][2] = 30;
	settings[6][3] = 6;
}

void ledprog_change(uint8_t idx)
{
	uint8_t i;

	ledprog = led_prog_list[idx];

	for (i = 0; i < 4; i++) {
		work[i] = 0;
	}
}

void ledprogi_twinkle(uint8_t r, uint8_t g, uint8_t b)
{
	uint8_t x;

	x = prng_get8();

	led_sk6x_set_all(0, 0, 0);

	if (x > 240) {
		x = prng_get32() & 0x7;
		led_sk6x_set(x, r, g, b);
	}
}

/*
 * settings:
 * 0	dwell
 * 1	threshold
 */
void ledprog0_twinkle_white()
{
	ledprogi_twinkle(0xff, 0xff, 0xff);
}

/*
 * settings:
 * 0	dwell
 * 1	threshold
 */
void ledprog1_twinkle_rgb_rand()
{
	uint32_t x;

	x = prng_get32();

	// set values
	r = (x >> 24);
	g = (x >> 16) & 0xff;
	b = (x >>  8) & 0xff;

	// randomly reduce colors to remove "whiteness"
	switch (x & 0x3) {
		case 1: {
			r >>= 4;
			g >>= 4;
			break;
		}
		case 2: {
			g >>= 4;
			b >>= 4;
			break;
		}
		case 3: {
			b >>= 4;
			r >>= 4;
			break;
		}
	}

	ledprogi_twinkle(r, g, b);
}


/*
 * settings:
 * 0	dwell
 * 1	threshold
 * 2	hue
 * 3	val
 */
void ledprog2_twinkle_rgb_set()
{
	hsv2rgb_8b(settings[2][2] * 6, 255, settings[2][3], &r, &g, &b);

	ledprogi_twinkle(r, g, b);
}

/*
 * settings:
 * 0	dwell
 * 1	threshold
 * 2	speed
 * 3	skip
 *
 * work:
 * 0	base hue
 */
void ledprog3_rainbow()
{
	uint8_t i;
	uint16_t w;

	work[0] += (settings[3][2] * 6) + 1;
	work[0] %= 1536;
	w = work[0];

	for (i = 0; i < 8; i++) {
		hsv2rgb_8b(w, 255, 255, &r, &g, &b);
		led_sk6x_set(i, r, g, b);

		w += (settings[3][3] * 6) + 1;
		w %= 1536;
	}
}

void ledprogi_flicker()
{
	uint16_t h;
	uint8_t x;

	h = settings[4][2] * 6;

	x = prng_get8();

	if (x > settings[4][1]) {
		x = 255;
	} else {
		x = settings[4][3];
	}

	x &= ~0x07;
	x |= prng_get32() & 0x07;

	hsv2rgb_8b(h, 255, x, &r, &g, &b);
}

/*
 * settings:
 * 0	dwell
 * 1	threshold
 * 2	hue
 * 3	base
 */
void ledprog4_flicker_same()
{
	ledprogi_flicker();
	led_sk6x_set_all(r, g, b);
}

/*
 * settings (uses ledprog4)
 * 0	dwell
 * 1	threshold
 * 2	hue
 * 3	val
 */
void ledprog5_flicker_all()
{
	uint8_t i;

	for (i = 0; i < 8; i++) {
		ledprogi_flicker();
		led_sk6x_set(i, r, g, b);
	}
}

/*
 * settings
 * 0	dwell
 * 1	speed
 * 2	hue
 * 3	faderate
 *
 * work:
 * 0	ramper state
 */
void ledprog6_ramper()
{
	uint8_t i;

	uint16_t h;
	int16_t spd;

	uint16_t floor = 0;
	uint8_t v[8];
	uint8_t fr;


	h = settings[6][2] * 6;
	fr = settings[6][3] & 0x3f;

	// speed -64 to -1, +1 to +64
	spd = (settings[6][1] >> 1) - 64;
	if (spd >= 0) spd++;
	work[0] += spd;
	work[0] &= 0xfff;

	// unpack fade values
	for (i = 0; i < 4; i++) {
		v[0 + i] = work[2] >> (i * 8);
		v[4 + i] = work[3] >> (i * 8);
	}

	for (i = 0; i < 8; i++) {
		// set to max if within range
		if (work[0] >= floor && work[0] <= floor + 64) {
			v[i] = 255;
		}

		// set LED
		hsv2rgb_8b(h, 255, v[i], &r, &g, &b);
		led_sk6x_set(i, r, g, b);

		// set fadeout value
		if (fr > v[i]) {
			v[i] = 0;
		} else {
			v[i] -= fr;
		}

		// both sides of top LEDs match
		if (i == 5) {
			i++;
			led_sk6x_set(i, r, g, b);
		}

		// work up the chain
		floor += 0x180;
	}

	// pack fade values
	work[2] = work[3] = 0;
	for (i = 0; i < 4; i++) {
		work[2] |= v[0 + i] << (i * 8);
		work[3] |= v[4 + i] << (i * 8);
	}
}
